/*
  ---- Delphes ----
  A Fast Simulator for general purpose LHC detector
  S. Ovyn ~~~~ severine.ovyn@uclouvain.be

  Center for Particle Physics and Phenomenology (CP3)
  Universite Catholique de Louvain (UCL)
  Louvain-la-Neuve, Belgium
*/

/// \file Delphes.cpp
/// \brief executable for the Delphes

#include "TChain.h"
#include "TApplication.h"
#include "TStopwatch.h"

#include "Utilities/ExRootAnalysis/interface/ExRootTreeReader.h"
#include "Utilities/ExRootAnalysis/interface/ExRootTreeWriter.h"
#include "Utilities/ExRootAnalysis/interface/ExRootTreeBranch.h"

#include "interface/DataConverter.h"
#include "interface/HEPEVTConverter.h"
#include "interface/LHEFConverter.h"
#include "interface/STDHEPConverter.h"

#include "interface/SmearUtil.h"
#include "interface/BFieldProp.h"
#include "interface/TriggerUtil.h"
#include "interface/VeryForward.h"
#include "interface/JetUtils.h"
#include "interface/FrogUtil.h"

#include <vector>
#include <iostream>

using namespace std;

//------------------------------------------------------------------------------
void todo(string filename) {
  ifstream infile(filename.c_str());
  cout << "** TODO list ..." << endl;
  while(infile.good()) {
    string temp;
    getline(infile,temp);
    cout << "*" << temp << endl;
  }
  cout << "** done...\n"; 
}

//------------------------------------------------------------------------------

int main(int argc, char *argv[])
{
  int appargc = 2;
  char *appName = "Delphes";
  char *appargv[] = {appName, "-b"};
  TApplication app(appName, &appargc, appargv);
  
  if(argc != 4 && argc != 3 && argc != 5) {
    cout << " Usage: " << argv[0] << " input_file output_file [detector_card] [trigger_card] " << endl;
    cout << " input_list - list of files in Ntpl, StdHep of LHEF format," << endl;
    cout << " output_file - output file." << endl;
    cout << " detector_card - Datacard containing resolution variables for the detector simulation (optional) "<<endl;
    cout << " trigger_card - Datacard containing the trigger algorithms (optional) "<<endl;
    exit(1);
  }
  
// 1. ********** initialisation ***********

  srand (time (NULL));         /* Initialisation du générateur */
  TStopwatch globalwatch, loopwatch, triggerwatch, frogwatch;
  globalwatch.Start();

  
  //read the input TROOT file
  string inputFileList(argv[1]), outputfilename(argv[2]);
  if(outputfilename.find(".root") > outputfilename.length() ) {
    cout << "output_file should be a .root file!\n"; 
    exit(1);
  }
  //create output log-file name
  string forLog = outputfilename;
  string LogName = forLog.erase(forLog.find(".root"));
  LogName = LogName+"_run.log";
  
  TFile *outputFile = TFile::Open(outputfilename.c_str(), "RECREATE"); // Creates the file, but should be closed just after
  outputFile->Close();
  
  string line;
  ifstream infile(inputFileList.c_str());
  infile >> line; // the first line determines the type of input files
  
  //read the datacard input file
  string DetDatacard("");
  if(argc>=4)  DetDatacard =argv[3];
  
  //Smearing information
  RESOLution *DET = new RESOLution();
  DET->ReadDataCard(DetDatacard);
  DET->Logfile(LogName);
  
  //read the trigger input file
  string TrigDatacard("data/trigger.dat");
  if(argc==5)  TrigDatacard =argv[4];
  
  //Trigger information
  TriggerTable *TRIGT = new TriggerTable();
  TRIGT->TriggerCardReader(TrigDatacard.c_str());
  TRIGT->PrintTriggerTable(LogName);
  
  //Propagation of tracks in the B field
  TrackPropagation *TRACP = new TrackPropagation(DetDatacard); 
  
  //Jet information
  JetsUtil *JETRUN = new JetsUtil(DetDatacard);
  
  //VFD information
  VeryForward * VFD = new VeryForward(DetDatacard);
 
  // data converters
  DataConverter *converter=0;
  
  if(strstr(line.c_str(),".hep"))
    {                           
      cout<<"#**********************************************************************"<<endl;
      cout<<"#**********         StdHEP file format detected           *************"<<endl;
      cout<<"#***********     Starting convertion to TRoot format     **************"<<endl;
      cout<<"#**********************************************************************"<<endl;
      converter = new STDHEPConverter(inputFileList,outputfilename);//case ntpl file in input list
    }
  else if(strstr(line.c_str(),".lhe"))
    {
      cout<<"#**********************************************************************"<<endl;
      cout<<"#***********          LHEF file format detected            ************"<<endl;
      cout<<"#***********     Starting convertion to TRoot format       ************"<<endl;
      cout<<"#**********************************************************************"<<endl;
      converter = new LHEFConverter(inputFileList,outputfilename);//case ntpl file in input list
    }
  else if(strstr(line.c_str(),".root"))
    {
      cout<<"#**********************************************************************"<<endl;
      cout<<"#**********         h2root file format detected           *************"<<endl;
      cout<<"#**********     Starting convertion to TRoot format       *************"<<endl;
      cout<<"#**********************************************************************"<<endl;
      converter = new HEPEVTConverter(inputFileList,outputfilename);//case ntpl file in input list
    }
  else { cout << "***  " << line.c_str() << "\n***  file format not identified\n***  Exiting\n"; return -1;};
  
  TChain chain("GEN");
  chain.Add(outputfilename.c_str());
  ExRootTreeReader *treeReader = new ExRootTreeReader(&chain);
  const TClonesArray *branchGen = treeReader->UseBranch("Particle");
  TIter itGen((TCollection*)branchGen);
  
  //Output file : contents of the analysis object data
  ExRootTreeWriter *treeWriter = new ExRootTreeWriter(outputfilename, "Analysis");
  ExRootTreeBranch *branchJet = treeWriter->NewBranch("Jet", TRootJet::Class());
  ExRootTreeBranch *branchTauJet = treeWriter->NewBranch("TauJet", TRootTauJet::Class());
  ExRootTreeBranch *branchElectron = treeWriter->NewBranch("Electron", TRootElectron::Class());
  ExRootTreeBranch *branchMuon = treeWriter->NewBranch("Muon", TRootMuon::Class());
  ExRootTreeBranch *branchPhoton = treeWriter->NewBranch("Photon", TRootPhoton::Class());
  ExRootTreeBranch *branchTracks = treeWriter->NewBranch("Tracks", TRootTracks::Class());
  ExRootTreeBranch *branchETmis = treeWriter->NewBranch("ETmis", TRootETmis::Class());
  ExRootTreeBranch *branchCalo = treeWriter->NewBranch("CaloTower", TRootCalo::Class());
  ExRootTreeBranch *branchZDC = treeWriter->NewBranch("ZDChits", TRootZdcHits::Class());
  ExRootTreeBranch *branchRP220 = treeWriter->NewBranch("RP220hits", TRootRomanPotHits::Class());
  ExRootTreeBranch *branchFP420 = treeWriter->NewBranch("FP420hits", TRootRomanPotHits::Class());
  
  TRootGenParticle *particle; 
  TRootETmis *elementEtmis;
  TRootElectron *elementElec;
  TRootMuon *elementMu;
  TRootPhoton *elementPhoton;
  TRootTracks *elementTracks;
  TRootCalo *elementCalo;
  
  TLorentzVector genMomentum(0,0,0,0);        // four-momentum at the vertex
  TLorentzVector genMomentumBfield(0,0,0,0);  // four-momentum at the exit of the tracks
  TLorentzVector momentumCaloSegmentation(0,0,0,0);    // four-momentum in the calo, after applying the calo segmentation
  LorentzVector jetMomentum;
  
  vector<fastjet::PseudoJet> input_particles;//for FastJet algorithm
  vector<fastjet::PseudoJet> sorted_jets;
  vector<TLorentzVector> TrackCentral;  
  vector<PhysicsTower> towers;
  vector<ParticleUtil> electron;
  vector<ParticleUtil> muon;
  vector<ParticleUtil> gamma;

  TSimpleArray<TRootGenParticle> NFCentralQ;
  float iPhi=0,iEta=0;



// 2. ********** Loop over all events ***********
  Long64_t entry, allEntries = treeReader->GetEntries();
  cout << "** The input list contains " << allEntries << " events" << endl;
  loopwatch.Start();

  // loop on all events
  for(entry = 0; entry < allEntries; ++entry)
    {
      TLorentzVector PTmis(0,0,0,0);
      treeReader->ReadEntry(entry);
      treeWriter->Clear();
      if((entry % 100) == 0 && entry > 0 )  cout << "** Processing element # " << entry << endl;
      
      electron.clear();
      muon.clear();
      gamma.clear();
      NFCentralQ.Clear();
      
      TrackCentral.clear();
      towers.clear();
      input_particles.clear();
      
      // 2.1 Loop over all particles in event
      itGen.Reset();
      while( (particle = (TRootGenParticle*) itGen.Next()) ) {
          int pid = abs(particle->PID);

	
	  // 2.1.1********************* preparation for the b-tagging
	  //// This subarray is needed for the B-jet algorithm
	  // optimization for speed : put first PID condition, then ETA condition, then either pt or status
	  if( (pid <= pB || pid == pGLUON) &&// is it a light quark or a gluon, i.e. is it one of these : u,d,c,s,b,g ?
              fabs(particle->Eta) < DET->CEN_max_tracker &&
              particle->Status != 1 &&
              particle->PT > DET->PT_QUARKS_MIN ) {
            NFCentralQ.Add(particle);
          }
	  
	  // 2.1.2 ********************* central detector: keep only visible particles
	  // keeps only final particles, visible by the central detector, including the fiducial volume
	  // the ordering of conditions have been optimised for speed : put first the STATUS condition
          if( (particle->Status == 1)    &&
	      ((pid != pNU1) && (pid != pNU2) && (pid != pNU3)) &&
              (fabs(particle->Eta) < DET->CEN_max_calo_fwd)
	      ) 
            { 
	      genMomentum.SetPxPyPzE(particle->Px, particle->Py, particle->Pz, particle->E);
	      genMomentumBfield = genMomentum;

	      // 2.1.2.1 ********************* central detector: magnetic field
	      // genMomentumBfield is then changed with respect to the magnetic field
	      if(DET->FLAG_bfield==1)  TRACP->Propagation(particle,genMomentumBfield);
	      float eta=fabs(genMomentumBfield.Eta());
	     
 
	      // 2.1.2.2 ********************* central detector: smearing (calorimeters, muon chambers)
	      switch(pid) {
		
	      case pE: // all electrons with eta < DET->MAX_CALO_FWD
		DET->SmearElectron(genMomentumBfield);
		if(genMomentumBfield.E()!=0 && eta < DET->CEN_max_tracker && genMomentumBfield.Pt() > DET->PTCUT_elec){
		  electron.push_back(ParticleUtil(genMomentumBfield,particle->PID));
		}
		break; // case pE
	      case pGAMMA: // all photons with eta < DET->MAX_CALO_FWD
		DET->SmearElectron(genMomentumBfield);
		if(genMomentumBfield.E()!=0 && eta < DET->CEN_max_tracker && genMomentumBfield.Pt() > DET->PTCUT_gamma) {
		  gamma.push_back(ParticleUtil(genMomentumBfield,particle->PID));
		}
		break; // case pGAMMA
	      case pMU: // all muons with eta < DET->MAX_MU
		DET->SmearMu(genMomentumBfield);
		if(genMomentumBfield.E()!=0 && eta < DET->CEN_max_mu &&  genMomentumBfield.Pt() > DET->PTCUT_muon){
		  muon.push_back(ParticleUtil(genMomentumBfield,particle->PID));
		}
		break; // case pMU
	      case pLAMBDA: // all lambdas with eta < DET->MAX_CALO_FWD
	      case pK0S:    // all K0s with eta < DET->MAX_CALO_FWD
		DET->SmearHadron(genMomentumBfield, 0.7);
		break; // case hadron
	      default:   // all other final particles with eta < DET->MAX_CALO_FWD
		DET->SmearHadron(genMomentumBfield, 1.0);
		break;
	      } // 2.1.2.2  switch (pid)
	     
 
	      // 2.1.2.3 ********************* central detector: calotowers
	      // all final particles but muons and neutrinos	
	      // for calorimetric towers and missing PT
	      int charge=Charge(pid); 
	      if(genMomentumBfield.E() !=0 && pid != pMU) {
		// in case the Bfield is not simulated, checks that charged particles have enough pt to reach the calos
		if ( !DET->FLAG_bfield && charge!=0 && genMomentumBfield.Pt() <= DET->TRACK_ptmin ) { /* particules do not reach calos */ }
		else { // particles reach calos	
	  	  // applies the calo segmentation and returns iEta & iPhi
		  DET->BinEtaPhi(genMomentumBfield.Phi(), genMomentumBfield.Eta(), iPhi, iEta); 
                  if(iEta != -100 && iPhi != -100) { 
                      momentumCaloSegmentation.SetPtEtaPhiE(genMomentumBfield.Pt(),iEta,iPhi,genMomentumBfield.E());
                      elementCalo = (TRootCalo*) branchCalo->NewEntry();
                      elementCalo->Set(momentumCaloSegmentation);
		      PhysicsTower Tower(LorentzVector(momentumCaloSegmentation.Px(),momentumCaloSegmentation.Py(),momentumCaloSegmentation.Pz(),momentumCaloSegmentation.E()));
                      towers.push_back(Tower);
                  } // if iEta != -100
		} // else : when particles reach the calos
	      } // 2.1.2.3 calotowers
	      

	      // 2.1.2.4 ********************* central detector: tracks
	      // all final charged particles
	      if(
		 (genMomentumBfield.E()!=0) &&
		 (fabs(genMomentumBfield.Eta()) < DET->CEN_max_tracker) && 
		 (DET->FLAG_bfield || ( !DET->FLAG_bfield && genMomentumBfield.Pt() > DET->TRACK_ptmin )) &&     
			// if bfield not simulated, pt should be high enough to be taken into account
		 ((rand()%100) < DET->TRACK_eff)  &&
		 (charge!=0)
		 )
		{
		  elementTracks = (TRootTracks*) branchTracks->NewEntry();
		  elementTracks->Set(genMomentum); // fills px,py,pz,pt,e,eta,phi at vertex
	 	  elementTracks->Etaout = genMomentumBfield.Eta();
	 	  elementTracks->Phiout = genMomentumBfield.Phi();
		  // TODO!!! apply a smearing on the position of the origin of the track
		  // elementTracks->SetPosition(particle->X,particle->Y,particle->Z);
		  // uses the output of the bfield computation : Xout=... Yout=... Zout...
		  // TODO!!! elementTrakcs->SetPositionOut(Xout,Yout,Zout);
		  TrackCentral.push_back(genMomentum); // tracks at vertex!
		} // 2.1.2.4 tracks
	      
	    } // 2.1.2 central detector
	  
	  // 2.1.3 ********************* forward detectors: zdc
	  if(DET->FLAG_vfd==1) {
	     VFD->ZDC(treeWriter,branchZDC,particle);
	     VFD->RomanPots(treeWriter,branchRP220,branchFP420,particle);
	  }
	  
       } // 2.1 while : loop on all particles of the event.
     


      // 2.2 ********** Output preparation & complex objects  ***********
 
      // 2.2.1 ********************* sorting collections by decreasing pt
      DET->SortedVector(electron); 
      for(unsigned int i=0; i < electron.size(); i++) {
        elementElec = (TRootElectron*) branchElectron->NewEntry();
        elementElec->Set(electron[i].Px(),electron[i].Py(),electron[i].Pz(),electron[i].E());
        elementElec->Charge = sign(electron[i].PID());
        elementElec->IsolFlag = DET->Isolation(electron[i].Phi(),electron[i].Eta(),TrackCentral,2.0);//isolation based on tracks
      }
      DET->SortedVector(muon);
      for(unsigned int i=0; i < muon.size(); i++) {
        elementMu = (TRootMuon*) branchMuon->NewEntry();
        elementMu->Charge = sign(muon[i].PID());
        elementMu->Set(muon[i].Px(),muon[i].Py(),muon[i].Pz(),muon[i].E());
        elementMu->IsolFlag = DET->Isolation(muon[i].Phi(),muon[i].Eta(),TrackCentral,2.0);
      }
      DET->SortedVector(gamma);
      for(unsigned int i=0; i < gamma.size(); i++) {
	elementPhoton = (TRootPhoton*) branchPhoton->NewEntry();
	elementPhoton->Set(gamma[i].Px(),gamma[i].Py(),gamma[i].Pz(),gamma[i].E());
      }
      
      // 2.2.2 ************* computes the Missing Transverse Momentum
      TLorentzVector Att(0.,0.,0.,0.);
      for(unsigned int i=0; i < towers.size(); i++)
        {
	  Att.SetPxPyPzE(towers[i].fourVector.px, towers[i].fourVector.py, towers[i].fourVector.pz, towers[i].fourVector.E);
	  if(fabs(Att.Eta()) < DET->CEN_max_calo_fwd)
	  {
            PTmis = PTmis + Att;
            // create a fastjet::PseudoJet with these components and put it onto
            // back of the input_particles vector
            input_particles.push_back(fastjet::PseudoJet(towers[i].fourVector.px,towers[i].fourVector.py,towers[i].fourVector.pz,towers[i].fourVector.E));
	  }
        }
      elementEtmis = (TRootETmis*) branchETmis->NewEntry();
      elementEtmis->ET = (PTmis).Pt();
      elementEtmis->Phi = (-PTmis).Phi();
      elementEtmis->Px = (-PTmis).Px();
      elementEtmis->Py = (-PTmis).Py();
      
      // 2.2.3 ************* B-tag, tau jets
      sorted_jets=JETRUN->RunJets(input_particles);
      JETRUN->RunJetBtagging(treeWriter, branchJet,sorted_jets,NFCentralQ);
      JETRUN->RunTauJets(treeWriter,branchTauJet,sorted_jets,towers, TrackCentral);
      
      treeWriter->Fill();
    } // 2. Loop over all events ('for' loop)
  
  treeWriter->Write();
  delete treeWriter;
  loopwatch.Stop();
 

 
// 3. ********** Trigger  & Frog ***********
  // 3.1 ************ running the trigger in case the FLAG trigger is put to 1 in the datacard
  triggerwatch.Start();
  if(DET->FLAG_trigger == 1)
    {
      // input
      TChain chainT("Analysis");
      chainT.Add(outputfilename.c_str());
      ExRootTreeReader *treeReaderT = new ExRootTreeReader(&chainT);
      
      // output
      TClonesArray *branchElecTrig = treeReaderT->UseBranch("Electron");
      TClonesArray *branchMuonTrig = treeReaderT->UseBranch("Muon");
      TClonesArray *branchJetTrig = treeReaderT->UseBranch("Jet");
      TClonesArray *branchTauJetTrig = treeReaderT->UseBranch("TauJet");
      TClonesArray *branchPhotonTrig = treeReaderT->UseBranch("Photon");
      TClonesArray *branchETmisTrig = treeReaderT->UseBranch("ETmis");
      
      ExRootTreeWriter *treeWriterT = new ExRootTreeWriter(outputfilename, "Trigger");
      ExRootTreeBranch *branchTrigger = treeWriterT->NewBranch("TrigResult", TRootTrigger::Class());
     
 
      Long64_t entryT, allEntriesT = treeReaderT->GetEntries();
      cout << "** Trigger: the 'Analysis' tree contains " << allEntriesT << " events" << endl;
      // loop on all entries
      for(entryT = 0; entryT < allEntriesT; ++entryT) {
          treeWriterT->Clear();
          treeReaderT->ReadEntry(entryT);
          TRIGT->GetGlobalResult(branchElecTrig, branchMuonTrig,branchJetTrig, branchTauJetTrig,branchPhotonTrig, branchETmisTrig,branchTrigger);
          treeWriterT->Fill();
      } // loop on all entries
      
      treeWriterT->Write();
      delete treeWriterT;
    } // trigger
    triggerwatch.Stop();
 
 
  // 3.2 ************** FROG display
  frogwatch.Start();
  if(DET->FLAG_frog == 1)
    {
      FrogDisplay *FROG = new FrogDisplay();
      FROG->BuidEvents(outputfilename,DET->NEvents_Frog);
      FROG->BuildGeom(DetDatacard);
    }
    frogwatch.Stop();
  



// 4. ********** End & Exit ***********
  cout << "** Exiting..." << endl;
  globalwatch.Stop();
  cout << "** Time report for " << allEntries << " events.\n";
  cout << " +  Time (s): \tCPU \t real"<< endl;
  cout << " +  Global:  \t" << globalwatch.CpuTime() << " \t " << globalwatch.RealTime() << endl;
  cout << " +  Events:  \t" << loopwatch.CpuTime() << " \t " << loopwatch.RealTime() << endl;
  if(DET->FLAG_trigger == 1)
  cout << " +  Trigger: \t" << triggerwatch.CpuTime() << " \t " << triggerwatch.RealTime() << endl;
  if(DET->FLAG_frog == 1)
  cout << " +  Frog:    \t" << frogwatch.CpuTime() << " \t " << frogwatch.RealTime() << endl;

  
  delete treeReader;
  delete DET;
  delete TRIGT;
  delete TRACP;
  delete JETRUN;
  delete VFD;
  if(converter) delete converter;
  
//  todo("TODO");
}
